/*
**    TP ETI5 CPE Lyon
**    Copyright (C) 2010 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _V3_HPP_
#define _V3_HPP_

#include <string>
#include <iostream>

namespace cpe
{

    /** \brief Vectors/Points 3D */
    class v3
    {
    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor */
        v3();
        /** \brief direct constructor */
        v3(const double& x,const double& y,const double& z);

        // ********************************************* //
        // ********************************************* //
        //  Get/set
        // ********************************************* //
        // ********************************************* //

        /** \brief get x coordinate */
        const double& x() const;
        /** \brief get x coordinate */
        double& x();
        /** \brief get y coordinate */
        const double& y() const;
        /** \brief get y coordinate */
        double& y();
        /** \brief get z coordinate */
        const double& z() const;
        /** \brief get z coordinate */
        double& z();

        /** \brief Access to the k_th entry (k in [0,2]) */
        const double& operator[](const size_t& k) const;
        /** \brief Access to the k_th entry (k in [0,2]) */
        double& operator[](const size_t& k);
        /** \brief Access to the k_th entry (k in [0,2]) */
        const double& operator()(const size_t& k) const;
        /** \brief Access to the k_th entry (k in [0,2]) */
        double& operator()(const size_t& k);

        /** \brief set every entry to 0 */
        void set_zero();


        // ********************************************* //
        // ********************************************* //
        //  Math operation
        // ********************************************* //
        // ********************************************* //

        /** \brief perform dot product between two v3 */
        double dot(const v3& p) const;

        /** \brief get the norm of the vector */
        double norm() const;
        /** \brief get the square norm of the vector */
        double norm2() const;
        /** \brief normalize the vector to unit length */
        v3 normalized() const;

        /** \brief compute cross product with an other vector */
        v3 cross(const v3& p) const;




        // ********************************************* //
        // ********************************************* //
        //  Operator +-*/
        // ********************************************* //
        // ********************************************* //

        /** \brief + operator */
        friend v3 operator+(const v3& p1,const v3& p2);
        /** \brief + operator */
        friend v3 operator+(const double& s,const v3& p);
        /** \brief + operator */
        friend v3 operator+(const v3& p,const double& s);

        /** \brief - operator */
        friend v3 operator-(const v3& p1,const v3& p2);
        /** \brief - operator */
        friend v3 operator-(const double& s,const v3& p);
        /** \brief - operator */
        friend v3 operator-(const v3& p,const double& s);

        /** \brief multiply by a scalar operator */
        friend v3 operator*(const double& s,const v3& p);
        /** \brief multiply by a scalar operator */
        friend v3 operator*(const v3& p,const double& s);
        /** \brief divide by a scalar operator */
        friend v3 operator/(const v3& p,const double& s);

        /** \brief internal + */
        v3& operator+=(const v3& p);
        /** \brief internal + */
        v3& operator+=(const double& s);
        /** \brief internal - */
        v3& operator-=(const v3& p);
        /** \brief internal - */
        v3& operator-=(const double& s);
        /** \brief internal * */
        v3& operator*=(const double& s);
        /** \brief internal / */
        v3& operator/=(const double& s);

        /** \brief unary negation */
        v3 operator-() const;

        /** \brief does componentwise mutliplication
            \return MC_v2d(x*p.x(),y*p.y(),z*p.z())
        */
        v3 product_compontentwise(const v3& p) const;
        /** \brief does componentwise mutliplication

            perform p=(p.x()*p2.x(),p.y()*p2.x())
        */
        v3& product_compontentwise_internal(const v3& p);


        // ********************************************* //
        // ********************************************* //
        //  Output
        // ********************************************* //
        // ********************************************* //

        /** \brief export the value as string cout<<v3(2,3,6) => 2 3 6 */
        std::string to_string() const;

        /** \brief output the vector in ostream as (x,y,z) */
        friend std::ostream& operator<<(std::ostream& stream,const v3& p);

    private:


        // ********************************************* //
        // ********************************************* //
        //  Internal parameters
        // ********************************************* //
        // ********************************************* //

        /** \brief x coordinate */
        double internal_x;
        /** \brief y coordinate */
        double internal_y;
        /** \brief z coordinate */
        double internal_z;


        // ********************************************* //
        // ********************************************* //
        //  Helper
        // ********************************************* //
        // ********************************************* //

        /** \brief assert that a size_t belongs to [[0,2]] */
        void assert_size(const size_t& k) const throw(std::exception);

    };

}



#endif
